package Risk.Internals;

import Risk.Jugador.AttackResult;
import Risk.Jugador.OccupaResult;
import Risk.Internals.Exceptions.*;
 
/**
* Classe principal del lloc, aqui es a on es construeixen les llistes que ens
* indiquen els territoris/continents que tenim. Aqui tenim funcions tan de consulta
* com de modificaci del mapa.
*/
public class Mapa {
  // Llista de territoris es una llista de llistes (Graf) que enllaa el territoris fronterisos
  // Llista de Continents igual pero enllaa continents fronterisos.
  // Llista de jugadors
  private Llista LTerritories, LContinents, LPlayers;
  private int torn; // ID del jugador que li toca moure
  private int armiesPendents; // Nombre d'armies pendents a colocar
  private final int TotalArmiesNoves = 20;  // Nombre d'armies que deixem colocar per cada volta
  
  public Mapa() throws NoMemoria, PDIdreta{
    LTerritories = new Llista();
    LContinents = new Llista();
    LPlayers = new Llista();
    torn = 0; // Torn de ningu
    armiesPendents=TotalArmiesNoves;
  }

  /**
  * Retorna el nombre d'armies que deixem ficar en cada placement de cada volta
  */
  public int armiesNum(){
    return TotalArmiesNoves;
  }

  /**
  * El Jugador ha ficat ja un army, el restem
  */
  public void restarArmiesPendents(){
    armiesPendents--;
  }
  
  /**
  * Comena un nou placement d'una nova volta. Cada jugador torna a tenir el maxim d'armies pendents.
  */
  public void resetArmiesPendents(){
    armiesPendents=TotalArmiesNoves;
  }
  
  
  /**
  * Estableix un torn inicial
  * @param n ID del jugador
  */
  public void torn(int n){
    if (n <= LPlayers.elements()){
      torn = n;
    }
  }
  
  /**
  * Pasa al seguent torn
  */
  public void tornSeguent() throws PDIdreta {
    Player p;
    
    if (torn != 0){ // Si encara no s'ha establer el torn inicial, no comencem
      // Reset al boolea que indica que hi ha hagut conquesta durant el torn
      LPlayers.principi();
      while (! LPlayers.finalLlista()) {
        p=(Player)LPlayers.actual();
        p.ResetConquest();
        LPlayers.avanzar();
      }
      
      if (torn+1 > LPlayers.elements()) {
        torn =1;
      } else {
        torn++;
      }
    }
  }
  
  /**
  * Retorna el torn
  * @return ID del jugador que te el torn
  */
  public int torn(){
    return torn;
  }
  
  /***************************** CREACI DEL MAPA *****************************/
  
  /************ JUGADORS ***************/
  
  /**
  * Afegeix un jugador al sistema
  * @param s Nom del jugador
  */
  public void afegirJugador(String s) throws NoMemoria, PDIdreta, JaExisteix{
    Player p;
    
    if (LPlayers.buida()){
      LPlayers.insertar(new Player(s, LPlayers.elements()+1));
    } else {
      LPlayers.principi();
      while (! LPlayers.finalLlista()){
        p = (Player)LPlayers.actual();
        if (s.equals(p.nom())) {
          throw new JaExisteix();
        }
        LPlayers.avanzar();
      }
      LPlayers.anarAlFinal();
      LPlayers.insertar(new Player(s, LPlayers.elements()+1));
    }
  }

  /**
  * Busca un jugador a partir del seu ID. Noms per s intern.
  * @param id ID del Jugador
  * @return Jugador
  */
  private Player buscarJugador(int id) throws JaExisteix, NoExisteix, PDIdreta {
    // Els jugadors es numeraran a partir de l'1, pero internament el tenim a partir del 0
    id = id-1;
    
    if(LPlayers.buida()){
      throw new NoExisteix();
    }
    
    LPlayers.principi();
    try {
      LPlayers.avanzar(id);
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    
    return (Player)LPlayers.actual();
  }
  /************ TERRITORIS ***************/
  
  /**
  * Busca si existeix un territori a la llista amb el mateix nom, si el troba retorna el
  * Territory sino null. Noms us intern.
  * @param s String del territori que busquem
  * @return Null si no existeix el territori o el territori en si.
  */
  private Territory buscarTerritori(String s) throws NoMemoria, PDIdreta{
    if (LTerritories.buida()){
      return null;  // Error: no existeix
    }
    LTerritories.principi(); // Cerca
    while ((!LTerritories.finalLlista()) && (!s.equals( ((Territory)LTerritories.actual()).nom()) )){
      LTerritories.avanzar();
    }
    if (!LTerritories.finalLlista()) {
      return (Territory)LTerritories.actual();
    } else {
      return null;  // Error: no existeix
    }
  }

  
  /**
  * Afegeix un Territori a la llista. Vigila que no hi estigues ja.
  * @param s String amb el nom del territori
  * @param c Integer que identifica el continent al que pertany
  */
  public void afegirTerritori(String s, int c) throws NoMemoria, JaExisteix, PDIdreta, NoExisteix {
    Territory t;
    Continent conti;

    // Comprovem que no existeix un altre territori amb el mateix nom
    if (! LTerritories.buida()) {
      LTerritories.principi();
      do {
        t = (Territory)LTerritories.actual();
        if (s.equals(t.nom())) {
          throw new JaExisteix();
        }
        LTerritories.avanzar();
      }while (! LTerritories.finalLlista());
      if (s.equals(t.nom())) {
          throw new JaExisteix();
      }
    }
    
    t = new Territory(s, LTerritories.elements()+1, c);
    LTerritories.anarAlFinal();
    LTerritories.insertar(t);
    
    
    LContinents.principi();
    try {
      LContinents.avanzar(c-1);
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    conti = (Continent)LContinents.actual();
    conti.afegirTerritory(t);
  }
  
  /**
  * Conecta 2 territoris que son frontera.
  * @param sTerritory1 String amb el nom del primer territori que es vol conectar
  * @param sTerritory2 String amb el nom del segon territori que es vol conectar
  */
  public void afegirTerritoryBorder(String sTerritory1, String sTerritory2) throws NoMemoria, PDIdreta, NodesIguals, NoExisteix, JaConectats{
    Territory Territory1, Territory2, nouInfo;

    if (sTerritory1.equals(sTerritory2)) { // No permitim enllaos amb el mateix, ni temps negatius
      throw new NodesIguals();
    }
    
    Territory1 = buscarTerritori(sTerritory1);  // Busquem node origen
    Territory2 = buscarTerritori(sTerritory2);
    if ((Territory1 == null) || (Territory2 == null)){   // No existeix algun dels territoris
      throw new NoExisteix();
    } else {    // Si existeix origen, podria haver-hi connexi
      
      if (Territory1.borderWith(sTerritory2)) {
        throw new JaConectats();
      } else {
        Territory1.borderWith(Territory2);
        Territory2.borderWith(Territory1);
      }
    }
  }

  
  /**************** CONTINENTS ****************/
  
  /**
  * Busca si existeix un continent a la llista amb el mateix nom, si el troba retorna el
  * Continent sino null. Noms s intern.
  * @param s String del continent que busquem
  * @return Null si no existeix el continent o el continent en si.
  */
  private Continent buscarContinent(String s) throws NoMemoria, PDIdreta{
    if (LContinents.buida()){
      return null;  // Error: no existeix
    }
    LContinents.principi(); // Cerca
    while ((!LContinents.finalLlista()) && (!s.equals( ((Continent)LContinents.actual()).nom()) )){
      LContinents.avanzar();
    }
    if (!LContinents.finalLlista()) {
      return (Continent)LContinents.actual();
    } else {
      return null;  // Error: no existeix
    }
  }
  

  /**
  * Afegeix un continent a la llista. Vigila que no hi estigues ja.
  * @param s String amb el nom del continent
  */
  public void afegirContinent(String s) throws NoMemoria, JaExisteix, PDIdreta {
    Continent t;

    // Comprovem que no existeix un altre continent i amb el mateix nom
    if (! LContinents.buida()) {
      LContinents.principi();
      do {
        t = (Continent)LContinents.actual();
        if (s.equals(t.nom())) {
          throw new JaExisteix();
        }
        LContinents.avanzar();
      }while (! LContinents.finalLlista());
      if (s.equals(t.nom())) {
          throw new JaExisteix();
      }
    }
    
    t = new Continent(s, LContinents.elements()+1);
    LContinents.anarAlFinal();
    LContinents.insertar(t);
  }
  
  /**
  * Conecta 2 continent is que son frontera.
  * @param sContinent1 String amb el nom del primer continent que es vol conectar
  * @param sContinent2 String amb el nom del segon continent que es vol conectar
  */
  public void afegirContinentBorder(String sContinent1, String sContinent2) throws NoMemoria, PDIdreta, NodesIguals, NoExisteix, JaConectats{
    Continent Continent1, Continent2, nouInfo;

    if (sContinent1.equals(sContinent2)) { // No permitim enllaos amb el mateix, ni temps negatius
      throw new NodesIguals();
    }
    
    Continent1 = buscarContinent(sContinent1);  // Busquem node origen
    Continent2 = buscarContinent(sContinent2);
    if ((Continent1 == null) || (Continent2 == null)){   // No existeix algun dels Continents
      throw new NoExisteix();
    } else {    // Si existeix origen, podria haver-hi connexi
      
      if (Continent1.borderWith(sContinent2)) {
        throw new JaConectats();
      } else {
        Continent1.borderWith(Continent2);
        Continent2.borderWith(Continent1);
      }
    }
  }
  
  /***************************** FI CREACI DEL MAPA *****************************/

  
  /***************************** MODIFICACI DEL MAPA *****************************/
  
  /**
  * Assigna un territori al jugador que te el torn.
  * @param territory ID del Territori
  */
  public void assign(int territory) throws NoExisteix, PDIdreta{
    Territory t =  (buscarTerritori(territory));
    t.owner(torn);
    t.armies(1);
  }
  
  /**
  * Aumenta en 1 el nombre d'armies d'un territori
  * @param ID del territori
  */
  public void place(int territory) throws NoExisteix, PDIdreta, NoTorn{
    Territory t =  (buscarTerritori(territory));

    if (t.owner() == torn){
      t.armies(t.armies()+1);
    } else {
      throw new NoTorn();
    }
  }
  
  /**
  * Ataca un territori des d'un altre. Es produeix conquesta si l'atacant t
  * ms armies.
  * @param ataca Tipus AttackResult que esta compost per un ID de territori origen i un altre de desti
  * @return Retorna True si s'ha produit conquesta.
  */
  public boolean attack(AttackResult ataca) throws NoExisteix, PDIdreta, JaExisteix, FromTo {
    Territory tFrom =  (buscarTerritori(ataca.GetFrom()));
    Territory tTo =  (buscarTerritori(ataca.GetTo()));
    Player p;
    if ((tFrom.owner() == torn) && (tTo.owner() != torn)) {
      if (tFrom.armies() > tTo.armies()) {
        if (tFrom.armies()-1 > 0) {
          tFrom.armies(tFrom.armies()-1);
        }
        tTo.owner(tFrom.owner());
        tTo.armies(1);
        p = buscarJugador(tFrom.owner());
        p.Conquest();
        return true;
      } else {
        if (tFrom.armies()-1 > 0) {
          tFrom.armies(tFrom.armies()-1);
        }
        if (tTo.armies()-2 > 0) {
          tTo.armies(tTo.armies()-2);
        }
        return false;
      }
    } else {
      throw new FromTo();
    }
  }
  
  /**
  * Realitza moviments d'armies entre territoris.
  * @parma ocupa Del tipus OccupaResult compost per un ID territori font, un desti i el nombre d'armies a moure.
  */
  public void occupation(OccupaResult ocupa) throws NoExisteix, PDIdreta, FromTo {
    Territory tFrom =  (buscarTerritori(ocupa.GetFrom()));
    Territory tTo =  (buscarTerritori(ocupa.GetTo()));
    int armies = ocupa.GetNumArmies();
    
    if ((tFrom.owner() == torn) && (tTo.owner() == torn)) {
      if (tFrom.armies() > armies) {
        tFrom.armies(tFrom.armies() - armies);
        
        tTo.armies(tTo.armies() + armies);
      }
    } else {
      throw new FromTo();
    }
    
  }
  
  /***************************** CONSULTES SOBRE EL MAPA *****************************/
  /************ JUGADORS ***************/
  
  /**
  * Retorna el nom del jugador amb l'id donat
  */
  public String PName(int id) throws JaExisteix, NoExisteix, PDIdreta {
    // Els jugadors es numeraran a partir de l'1, pero internament el tenim a partir del 0
    id = id-1;
    
    if(LPlayers.buida()){
      throw new NoExisteix();
    }
    
    LPlayers.principi();
    try {
      LPlayers.avanzar(id);
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    
    return ((Player)LPlayers.actual()).nom();
  }
  
  /**
  * Retorna True si el jugador indicat esta actiu (t el torn)
  */
  public boolean PActive(int id){
    return (id==torn);
  }
  
  /**
  * Retorna el nombre d'armies que te el jugador indicat
  */
  public int PArmiesCount(int id) throws NoExisteix, PDIdreta{
    Territory t;
    int total = 0;
    
    if (LTerritories.buida()) {
      return 0;
    }
    
    LTerritories.principi();
    while (! LTerritories.finalLlista()){
      t = (Territory)LTerritories.actual();
      if (t.owner() == id) {
        total += t.armies();
      }
      LTerritories.avanzar();
    }
    
    return total;
  }
  
  /**
  * Retorna el nombre d'armies del jugador indicat que encara ha de colocar
  */
  public int PNewArmies(int id){
    return armiesPendents;
  }
  
  /**
  * Retorna el nombre de territoris controlats pel jugador indicat.
  */
  public int PTerritories(int id) throws NoExisteix, PDIdreta{
    Territory t;
    int total = 0;
    
    if (LTerritories.buida()) {
      return 0;
    }
    
    LTerritories.principi();
    while (! LTerritories.finalLlista()){
      t = (Territory)LTerritories.actual();
      if (t.owner() == id) {
        total++;
      }
      LTerritories.avanzar();
    }
    
    return total;
  }
  
  /**************** TERRITORIS ****************/
  
  /**
  * Busca un territori a partir del seu ID. Noms per s intern.
  * @param id ID del territori
  * @return Territory
  */
  private Territory buscarTerritori(int id) throws NoExisteix, PDIdreta{
    // Els territoris van de 1 a 42, i nosaltres els tenim a la llista de 0 a 41
    id = id-1;
    
    if (LTerritories.buida()){
      throw new NoExisteix();
    }
    
    LTerritories.principi();
    try {
      LTerritories.avanzar(id);
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    return (Territory)LTerritories.actual();
  }
  
  /**
  * Retorna el nom del territori en la posici donada.
  * @param pos La posici del territori o el ID
  * @return Retorna el nomb del territori
  */
  public String TName(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).nom();
  }
  
  /**
  * Retorna l'id del propietari del territori indicat
  */
  public int TOwner(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).owner();
  }
  
  /**
  * Retorna el nombre d'armies del territori indicat
  */
  public int TArmies(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).armies();
  }
  
  /**
  * Retorna l'id del continent al qual pertany el territori indicat
  */
  public int TContinent(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).continentID();
  }
  
  /**
  * Retorna el nombre de fronteres que te el territori indicat
  */
  public int TBordersCount(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).borderTotal();
  }
  
  /**
  * Retorna l'id del territori numero "num" que fa frontera amb el territori indicat (id)
  */
  public int TBorder(int id, int num) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).Border(num);
  }
  
  /**
  * Retorna true si id1 (territori) fa frontera amb id2 (territori)
  */
  public boolean TIsBordering(int id1, int id2) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id1)).borderWith( this.TName(id2) );
  }
  
  /**
  * Retorna si el territori indicat fa frontera amb un territori enemic
  */
  public boolean TIsFront(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).isFront();
  }
  
  /**
  * Retorna True si el territori indicat ens pertany
  */
  public boolean TIsMine(int id) throws PDIdreta, NoExisteix{
    return ((buscarTerritori(id)).owner() == torn); // Sabem qui som pel torn
  }
  
  /**
  * Retorna el nombre de fronteres a territoris ocupas per enemics
  */
  public int TFrontsCount(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).FrontsCount();
  }
  
  /**
  * Retorna l'id del territori numero "num" que fa frontera amb el territori indicat (id) i es de l'enemic
  */
  public int TFront(int id, int num) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).Front(num);
  }
  
  /**
  * Retorna la frontera ocupada per enemics mes forta
  */
  public int TStrongestFront(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).StrongestFront();
  }
  
  /**
  * Retorna el nombre d'armis de la frontera ocupada per enemics mes forta
  */
  public int TStrongestFrontArmy(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).StrongestFrontArmy();
  }
  
  /**
  * Retorna la frontera ocupada per enemics mes debil
  */
  public int TWeakestFront(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).WeakestFront();
  }
  
  /**
  * Retorna el nombre d'armis de la frontera ocupada per enemics mes debil
  */
  public int TWeakestFrontArmy(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).WeakestFrontArmy();
  }
  
  /**
  * Retorna el total d'armies del territoris que fan frontera amb l'indicat
  */
  public int TPressure(int id) throws PDIdreta, NoExisteix{
    return (buscarTerritori(id)).Pressure();
  }
  
  /**************** CONTINENTS ****************/
  
  /**
  * Mtode privat que busca un continent i el retorna, sino retorna null
  * @return Continent si existeix, null si no.
  */
  private Continent buscarContinent(int id) throws NoExisteix, PDIdreta{
    // Els territoris van de 1 a 6, i nosaltres els tenim a la llista de 0 a 5
    id = id-1;
    
    if (LContinents.buida()){
      throw new NoExisteix();
    }
    
    LContinents.principi();
    try {
      LContinents.avanzar(id);
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    return (Continent)LContinents.actual();
  }
  
  /**
  * Retorna el nom del Continent en la posici donada.
  * @param pos La posici del Continent o el ID
  * @return Retorna el nomb del Continent
  */
  public String CName(int pos) throws NoMemoria, PDIdreta, NoExisteix{
    // Els Continents van de 1 a 6, i nosaltres els tenim a la llista de 0 a 5
    pos = pos-1;
    
    if (LContinents.buida()){
      throw new NoExisteix();
    }
    LContinents.principi();
    try {
      LContinents.avanzar(pos);   // Ens situem a la posici
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    return ((Continent)LContinents.actual()).nom();
  }
  
  /**
  * Metode per saber el propietari d'un terreny
  * @return Retorna 0 si te ms d'un ocupant, sino l'id de l'ocupant
  */
  public int COwner(int id) throws PDIdreta{
    Territory t;
    int own = 0;
    
    if (LTerritories.buida()) {
      return 0;
    }
    
    LTerritories.principi();
    do {
      t = (Territory)LTerritories.actual();
      if ((own==0) && (t.continentID() == id)){    // Si encara no s'havia assignat cap owner
        own = t.owner();
      }else if ((t.owner() != own) && (t.continentID() == id)) {
        return 0;
      }
      LTerritories.avanzar();
    } while (! (LTerritories.finalLlista()));
    
    return own;
  }
  
  /**
  * Nombre de territoris que te un continent
  * @param id ID del continent que volem saber quants territoris te
  * @return Retorna el nombre de territoris que pertanyen a un continent
  */
  public int CTerritoriesCount(int id) throws PDIdreta{
    Territory t;
    int total = 0;
    
    if (LTerritories.buida()) {
      return 0;
    }
    
    LTerritories.principi();
    do {
      t = (Territory)LTerritories.actual();
      if (t.continentID() == id){
        total++;
      }
      LTerritories.avanzar();
    } while (! (LTerritories.finalLlista()));
    
    return total;
  }
  
  
  /**
  * Retorna l'ID del territori numero "num" del continent indicat (id)
  * @param id ID del continent
  * @param num Posici del territori dintre del continent
  */
  public int CTerritory(int id, int num) throws PDIdreta {
    Territory t;
    int index = 0;
    
    if (LTerritories.buida()){
      return 0;
    }
    
    LTerritories.principi();
    while (! LTerritories.finalLlista()){
      t = (Territory)LTerritories.actual();
      if (t.continentID() == id){
        index++;
        if(index==num){
          return t.id();
        }
      }
      LTerritories.avanzar();
    }
    
    return 0;
  }
  
  
  /**
  * Retorna el nombre de territoris que rodejen un Continent
  * @param id ID del continent que volem saber quantes fronteres te
  * @return Retorna el nombre de territoris que fan frontera amb un continent
  */
  public int CBordersCount(int id) throws PDIdreta, NoExisteix{
    return (buscarContinent(id)).BorderCount();
  }

  /**
  * Rertorna l'id del territori numero "num" que fa frontera amb el continent indicat (id)
  * @param id ID del continent
  * @param num Posici que ocupa el territori dintre dels territoris que fan frontera amb el continent indicat
  * @return ID del territori que busquem
  */
  public int CBorder(int id, int num) throws PDIdreta, NoExisteix {
    return (buscarContinent(id)).Border(num);
  }
  
  /**************** STATUS DEL JOC ****************/
  /**
  * Retorna True si el jugador que te el torn (el que fa la crida) ha conquistat
  * algun territori durant el seu torn
  */
  public boolean SConquest() throws JaExisteix, NoExisteix, PDIdreta {
    return (buscarJugador(torn)).GetConquest();
  }
  
  /**
  * Retorna el nombre de jugadors
  */
  public int SPlayersCount(){
    return LPlayers.elements();
  }
  
  /***************************** FI CONSULTES SOBRE EL MAPA *****************************/
}